from datetime import datetime, timedelta
from typing import Optional
from lineage.bigquery_query import BigQueryQuery
from lineage.exceptions import SerializationError
from lineage.query import Query
from lineage.snowflake_query import SnowflakeQuery
from lineage.utils import is_flight_mode_on
import json
import os


class QueryHistory(object):

    INFORMATION_SCHEMA_QUERY_HISTORY = None

    LATEST_QUERY_HISTORY_FILE = './latest_query_history.json'

    def __init__(self, con, should_export_query_history: bool = True, ignore_schema: bool = False) -> None:
        self._con = con
        self._should_export_query_history = should_export_query_history
        self._ignore_schema = ignore_schema

    def _serialize_query_history(self, queries: [str]) -> None:
        if self._should_export_query_history:
            with open(self.LATEST_QUERY_HISTORY_FILE, 'w') as query_history_file:
                serialized_queries = []
                for query in queries:
                    serialized_queries.append(query.to_dict())
                json.dump(serialized_queries, query_history_file)

    def _deserialize_query_history(self) -> [Query]:
        deserialized_queries = []
        if os.path.exists(self.LATEST_QUERY_HISTORY_FILE):
            with open(self.LATEST_QUERY_HISTORY_FILE, 'r') as query_history_file:
                queries = json.load(query_history_file)
                for query_dict in queries:
                    platform_type = query_dict.pop('platform_type')
                    if platform_type == SnowflakeQuery.PLATFORM_TYPE:
                        deserialized_queries.append(SnowflakeQuery.from_dict(query_dict))
                    elif platform_type == BigQueryQuery.PLATFORM_TYPE:
                        deserialized_queries.append(BigQueryQuery.from_dict(query_dict))
                    else:
                        raise SerializationError(f'Invalid platform type - {platform_type}')

        return deserialized_queries

    @staticmethod
    def _include_end_date(end_date: datetime) -> Optional[datetime]:
        if end_date is not None and (end_date.hour, end_date.minute, end_date.second) == (0, 0, 0):
            return end_date + timedelta(hours=23, minutes=59, seconds=59)

        return end_date

    def extract_queries(self, start_date: datetime, end_date: datetime) -> [Query]:
        if is_flight_mode_on():
            queries = self._deserialize_query_history()
        else:
            queries = self._query_history_table(start_date, end_date)

            self._serialize_query_history(queries)

        return queries

    def _query_history_table(self, start_date: datetime, end_date: datetime) -> [Query]:
        pass

    def get_database_name(self):
        pass

    def get_schema_name(self):
        pass



from datetime import datetime
from typing import Optional, Union
import dateutil.parser


class QueryContext(object):
    def __init__(self, queried_database: Optional[str] = None, queried_schema: Optional[str] = None,
                 query_time: Optional[datetime] = None, query_volume: Optional[int] = None,
                 query_type: Optional[str] = None, user_name: Optional[str] = None,
                 role_name: Optional[str] = None, referenced_tables: [dict] = None,
                 destination_table: dict = None) -> None:
        self.queried_database = queried_database
        self.queried_schema = queried_schema
        self.query_time = query_time
        self.query_volume = query_volume
        self.query_type = query_type
        self.user_name = user_name
        self.role_name = role_name
        self.referenced_tables = referenced_tables if referenced_tables is not None else []
        self.destination_table = destination_table

    def to_dict(self) -> dict:
        return {'queried_database': self.queried_database,
                'queried_schema': self.queried_schema,
                'query_time': self._query_time_to_str(self.query_time),
                'query_volume': self.query_volume,
                'query_type': self.query_type,
                'user_name': self.user_name,
                'role_name': self.role_name,
                'referenced_tables': self.referenced_tables,
                'destination_table': self.destination_table}

    @staticmethod
    def _query_time_to_str(query_time: Optional[datetime], fmt: str = None) -> Optional[str]:
        if query_time is None:
            return None

        if fmt is None:
            return query_time.isoformat()

        return query_time.strftime(fmt)

    @staticmethod
    def _html_param_with_default(param: Union[str, int], default: Union[str, int] = 'Unknown') -> Union[str, int]:
        return default if param is None else param

    def to_html(self) -> str:
        query_type = self._html_param_with_default(self.query_type)
        user_name = self._html_param_with_default(self.user_name)
        role_name = self._html_param_with_default(self.role_name)
        query_time = self._query_time_to_str(self.query_time, fmt='%Y-%m-%d %H:%M:%S')
        query_volume = self._html_param_with_default(self.query_volume, 0)
        volume_color = "DarkSlateGrey"
        if query_volume == 0:
            volume_color = "tomato"

        return f"""
                    <html>
                        <body>
                            <div style="font-family:arial;color:DarkSlateGrey;font-size:110%;">
                                <strong>
                                    Last update</br>
                                </strong>
                                <div style="min-width:62px;display:inline-block">Type:</div> {query_type}</br>
                                <div style="min-width:62px;display:inline-block">User:</div> {user_name}</br>
                                <div style="min-width:62px;display:inline-block">Role:</div> {role_name}</br>
                                <div style="min-width:62px;display:inline-block">Time:</div> {query_time}</br>
                                <div style="min-width:62px;display:inline-block;">Volume:</div> <a style="color:{volume_color}">{query_volume} rows</a>
                            </div>
                        </body>
                    </html>
        """

    @staticmethod
    def from_dict(query_context_dict: dict) -> 'QueryContext':
        if 'query_time' in query_context_dict and query_context_dict['query_time'] is not None:
            query_context_dict['query_time'] = dateutil.parser.parse(query_context_dict['query_time'])
        return QueryContext(**query_context_dict)



import sqlparse
from sqllineage.core import LineageAnalyzer, LineageResult
from lineage.table_resolver import TableResolver
from lineage.query_context import QueryContext
from lineage.utils import get_logger

logger = get_logger(__name__)


class Query(object):
    PLATFORM_TYPE = None

    def __init__(self, raw_query_text: str, query_context: QueryContext, profile_database_name: str,
                 profile_schema_name: str) -> None:
        self._raw_query_text = raw_query_text
        self._query_context = query_context
        self._profile_database_name = profile_database_name
        self._profile_schema_name = profile_schema_name
        self.dropped_tables = set()
        self.renamed_tables = set()
        self.source_tables = set()
        self.target_tables = set()

    def to_dict(self) -> dict:
        return {'raw_query_text': self._raw_query_text,
                'query_context': self._query_context.to_dict(),
                'profile_database_name': self._profile_database_name,
                'profile_schema_name': self._profile_schema_name,
                'platform_type': self.PLATFORM_TYPE}

    @staticmethod
    def _query_text_to_analyzed_sql_statements(query_text: str) -> [LineageResult]:
        parsed_query = sqlparse.parse(query_text.strip())
        analyzed_statements = [LineageAnalyzer().analyze(statement) for statement in parsed_query
                               if statement.token_first(skip_cm=True, skip_ws=True)]
        return analyzed_statements

    def get_context_as_html(self) -> str:
        return self._query_context.to_html()

    @classmethod
    def _parse_query_text(cls, table_resolver: TableResolver, raw_query_text: str) -> (set, set, set, set):
        source_tables = set()
        target_tables = set()
        renamed_tables = set()
        dropped_tables = set()

        analyzed_statements = cls._query_text_to_analyzed_sql_statements(raw_query_text)
        for analyzed_statement in analyzed_statements:
            # Handle drop tables, if they exist in the statement
            for dropped_table in analyzed_statement.drop:
                dropped_tables.add(table_resolver.name_qualification(dropped_table))

            # Handle rename tables
            for old_table, new_table in analyzed_statement.rename:
                old_table_name = table_resolver.name_qualification(old_table)
                new_table_name = table_resolver.name_qualification(new_table)
                renamed_tables.add((old_table_name, new_table_name))

            # sqllineage lib marks CTEs as intermediate tables. Remove CTEs (WITH statements) from the source
            # tables.
            if not source_tables:
                source_tables = {table_resolver.name_qualification(source)
                                 for source in analyzed_statement.read - analyzed_statement.intermediate}
            elif len(analyzed_statement.read) > 0:
                logger.debug(f"Unexpected case when source_tables is already filled. Query -\n{raw_query_text}\n")

            if not target_tables:
                target_tables = {table_resolver.name_qualification(target) for target in analyzed_statement.write}
            elif len(analyzed_statement.write) > 0:
                logger.debug(f"Unexpected case when target_tables is already filled. Query -\n{raw_query_text}\n")

        return source_tables, target_tables, renamed_tables, dropped_tables

    def _get_platform_type(self) -> str:
        pass

    def parse(self, full_table_names: bool = False) -> None:
        pass






import itertools
from typing import Optional
import networkx as nx
from lineage.exceptions import ConfigError
from pyvis.network import Network
import webbrowser
from lineage.query import Query
from lineage.utils import get_logger
from tqdm import tqdm
import pkg_resources

logger = get_logger(__name__)

GRAPH_VISUALIZATION_OPTIONS = """{
            "edges": {
                "color": {
                  "color": "rgba(23,107,215,1)",
                  "highlight": "rgba(23,107,215,1)",
                  "hover": "rgba(23,107,215,1)",
                  "inherit": false
                },
                "dashes": true,
                "smooth": {
                    "type": "continuous",
                    "forceDirection": "none"
                }
            },
            "layout": {
                "hierarchical": {
                    "enabled": true,
                    "levelSeparation": 485,
                    "nodeSpacing": 100,
                    "treeSpacing": 100,
                    "blockShifting": false,
                    "edgeMinimization": true,
                    "parentCentralization": false,
                    "direction": "LR",
                    "sortMethod": "directed"
                }
            },
            "interaction": {
                "hover": true,
                "navigationButtons": true,
                "multiselect": true,
                "keyboard": {
                    "enabled": true
                }
            },
            "physics": {
                "enabled": false,
                "hierarchicalRepulsion": {
                    "centralGravity": 0
                },
                "minVelocity": 0.75,
                "solver": "hierarchicalRepulsion"
            }
}"""


class LineageGraph(object):
    UPSTREAM_DIRECTION = 'upstream'
    DOWNSTREAM_DIRECTION = 'downstream'
    BOTH_DIRECTIONS = 'both'
    SELECTED_NODE_COLOR = '#0925C7'
    SELECTED_NODE_TITLE = 'Selected table<br/>'

    def __init__(self, show_isolated_nodes: bool = False, show_full_table_names: bool = False) -> None:
        self._lineage_graph = nx.DiGraph()
        self._show_isolated_nodes = show_isolated_nodes
        self._show_full_table_names = show_full_table_names

    def _update_lineage_graph(self, query: Query) -> None:
        query.parse(self._show_full_table_names)

        # Handle drop tables, if they exist in the statement
        for dropped_table_name in query.dropped_tables:
            self._remove_node(dropped_table_name)

        # Handle rename tables
        for old_table_name, new_table_name in query.renamed_tables:
            self._rename_node(old_table_name, new_table_name)

        self._add_nodes_and_edges(query.source_tables, query.target_tables, query.get_context_as_html())

    def _add_nodes_and_edges(self, sources: {str}, targets: {str}, query_context_html: str) -> None:
        if None in sources:
            sources.remove(None)
        if None in targets:
            targets.remove(None)

        if not sources and not targets:
            return

        if len(sources) > 0 and len(targets) == 0:
            if self._show_isolated_nodes:
                self._lineage_graph.add_nodes_from(sources)
        elif len(targets) > 0 and len(sources) == 0:
            if self._show_isolated_nodes:
                self._lineage_graph.add_nodes_from(targets, title=query_context_html)
        else:
            self._lineage_graph.add_nodes_from(sources)
            self._lineage_graph.add_nodes_from(targets, title=query_context_html)
            for source, target in itertools.product(sources, targets):
                self._lineage_graph.add_edge(source, target)

    def _rename_node(self, old_node: str, new_node: str) -> None:
        if old_node is None or new_node is None:
            return

        if self._lineage_graph.has_node(old_node):
            # Rename in place instead of copying the entire lineage graph
            nx.relabel_nodes(self._lineage_graph, {old_node: new_node}, copy=False)

    def _remove_node(self, node: str) -> None:
        # First let's check if the node exists in the graph
        if node is not None and self._lineage_graph.has_node(node):
            node_successors = list(self._lineage_graph.successors(node))
            node_predecessors = list(self._lineage_graph.predecessors(node))

            # networknx's remove_node already takes care of in and out edges
            self._lineage_graph.remove_node(node)

            # Now that we have just deleted the dropped table from the graph, we need to take care of
            # new island nodes.
            if not self._show_isolated_nodes:
                for successor in node_successors:
                    if self._lineage_graph.has_node(successor) and self._lineage_graph.degree(successor) == 0:
                        self._lineage_graph.remove_node(successor)
                for predecessor in node_predecessors:
                    if self._lineage_graph.has_node(predecessor) and self._lineage_graph.degree(predecessor) == 0:
                        self._lineage_graph.remove_node(predecessor)

    def init_graph_from_query_list(self, queries: [Query]) -> None:
        logger.debug(f'Loading {len(queries)} queries into the lineage graph')
        for query in tqdm(queries, desc="Updating lineage graph", colour='green'):
            self._update_lineage_graph(query)

        logger.debug(f'Finished updating lineage graph!')

    def filter_on_table(self, selected_table: str, direction: str = None, depth: int = None) -> None:
        logger.debug(f'Filtering lineage graph on table - {selected_table}')

        if direction == self.DOWNSTREAM_DIRECTION:
            self._lineage_graph = self._downstream_graph(selected_table, depth)
        elif direction == self.UPSTREAM_DIRECTION:
            self._lineage_graph = self._upstream_graph(selected_table, depth)
        elif direction == self.BOTH_DIRECTIONS:
            downstream_graph = self._downstream_graph(selected_table, depth)
            upstream_graph = self._upstream_graph(selected_table, depth)
            self._lineage_graph = nx.compose(upstream_graph, downstream_graph)
        else:
            raise ConfigError(f'Direction must be one of the following - {self.UPSTREAM_DIRECTION}|'
                              f'{self.DOWNSTREAM_DIRECTION}|{self.BOTH_DIRECTIONS}, '
                              f'Got - {direction} instead.')

        self._update_selected_node_attributes(selected_table)
        logger.debug(f'Finished filtering lineage graph on table - {selected_table}')
        pass

    def _downstream_graph(self, source_node: str, depth: Optional[int]) -> nx.DiGraph:
        logger.debug(f'Building a downstream graph for - {source_node}, depth - {depth}')
        return nx.bfs_tree(G=self._lineage_graph, source=source_node, depth_limit=depth)

    def _upstream_graph(self, target_node: str, depth: Optional[int]) -> nx.DiGraph:
        logger.debug(f'Building an upstream graph for - {target_node}, depth - {depth}')
        reversed_lineage_graph = self._lineage_graph.reverse(copy=True)
        return nx.bfs_tree(G=reversed_lineage_graph, source=target_node, depth_limit=depth).reverse(copy=False)

    def _update_selected_node_attributes(self, selected_node: str) -> None:
        if self._lineage_graph.has_node(selected_node):
            node = self._lineage_graph.nodes[selected_node]
            node_title = node.get('title', '')
            node.update({'color': self.SELECTED_NODE_COLOR,
                         'title': self.SELECTED_NODE_TITLE + node_title})

    @staticmethod
    def _load_header() -> str:
        header_content = ""
        header = pkg_resources.resource_filename(__name__, "header.html")

        if header is not None:
            with open(header, 'r') as header_file:
                header_content = header_file.read()

        return header_content

    def draw_graph(self, should_open_browser: bool = True) -> None:
        # Visualize the graph
        net = Network(height="100%", width="100%", directed=True, heading=self._load_header())
        net.from_nx(self._lineage_graph)
        net.set_options(GRAPH_VISUALIZATION_OPTIONS)

        net.save_graph("elementary_lineage.html")
        if should_open_browser:
            webbrowser.open_new_tab('elementary_lineage.html')



from lineage.query_context import QueryContext
from sqllineage.exceptions import SQLLineageException
from lineage.table_resolver import TableResolver
from lineage.query import Query
from lineage.utils import get_logger

logger = get_logger(__name__)


class SnowflakeQuery(Query):
    PLATFORM_TYPE = 'SNOWFLAKE'

    @staticmethod
    def from_dict(query_dict: dict):
        query_context = QueryContext.from_dict(query_dict.pop('query_context'))
        if 'platform_type' in query_dict:
            query_dict.pop('platform_type')
        return SnowflakeQuery(**query_dict, query_context=query_context)

    def parse(self, full_table_names: bool = False) -> None:
        try:
            table_resolver = TableResolver(self._profile_database_name, self._profile_schema_name,
                                           self._query_context.queried_database, self._query_context.queried_schema,
                                           full_table_names)

            self.source_tables, self.target_tables, self.renamed_tables, self.dropped_tables = \
                self._parse_query_text(table_resolver, self._raw_query_text)
        except SQLLineageException as exc:
            logger.debug(f'SQLLineageException was raised while parsing this query -\n{self._raw_query_text}\n'
                         f'Error was -\n{exc}.')



from sqllineage.models import Schema, Table
from typing import Optional, Union
from lineage.utils import get_logger

logger = get_logger(__name__)


class TableResolver(object):

    def __init__(self, profile_database_name: str, profile_schema_name: str, queried_database_name: str = None,
                 queried_schema_name: str = None, full_table_names: bool = False) -> None:
        self._profile_database_name = profile_database_name
        self._profile_schema_name = profile_schema_name
        self._queried_database_name = queried_database_name
        self._queried_schema_name = queried_schema_name
        self._show_full_table_name = full_table_names

    @staticmethod
    def _resolve_table_qualification(table: Table, database_name: str, schema_name: str) -> Table:
        if not table.schema:
            if database_name is not None and schema_name is not None:
                table.schema = Schema(f'{database_name}.{schema_name}')
        else:
            parsed_query_schema_name = str(table.schema)
            if '.' not in parsed_query_schema_name:
                # Resolved schema is either empty or fully qualified with db_name.schema_name
                if database_name is not None:
                    table.schema = Schema(f'{database_name}.{parsed_query_schema_name}')
                else:
                    table.schema = Schema()
        return table

    def _should_ignore_table(self, table: Table) -> bool:
        if self._profile_schema_name is not None:
            if str(table.schema) == str(Schema(f'{self._profile_database_name}.{self._profile_schema_name}')):
                return False
        else:
            if str(Schema(self._profile_database_name)) in str(table.schema):
                return False

        return True

    def name_qualification(self, table: Union[Table, str]) -> Optional[str]:
        if isinstance(table, str):
            table = Table(table)

        # If queried database and schema names exist, prefer them when resolving the table name
        database_name = self._queried_database_name if self._queried_database_name is not None else \
            self._profile_database_name
        schema_name = self._queried_schema_name if self._queried_schema_name is not None else self._profile_schema_name

        table = self._resolve_table_qualification(table, database_name, schema_name)

        if self._should_ignore_table(table):
            return None

        if self._show_full_table_name:
            resolved_table_name = str(table)
        else:
            resolved_table_name = str(table).rsplit('.', 1)[-1]

        logger.debug(f'Resolved table name - {resolved_table_name}')
        return resolved_table_name



from typing import Dict, Any
import dbt.config
from dbt.context.base import generate_base_context
from dbt.exceptions import DbtConfigError
from dbt.adapters.bigquery.connections import BigQueryConnectionManager
import google.cloud.bigquery
import google.cloud.exceptions
from google.api_core import client_info
from lineage.exceptions import ConfigError
from lineage.utils import get_logger

logger = get_logger(__name__)


def extract_profile_data(profiles_raw: Dict[str, Any], profile_name: str, target_name: str) -> Dict[str, Any]:
    profile_data = dict()
    try:
        selected_profile = profiles_raw[profile_name]
        profile_data = selected_profile['outputs'][target_name]
    except KeyError as exc:
        logger.debug(f"Failed extracting profile data: {profiles_raw}, {profile_name}, {target_name}, {exc}")

    return profile_data


def extract_credentials_and_data_from_profiles(profiles_dir: str, profile_name: str):
    try:
        profiles_raw = dbt.config.profile.read_profile(profiles_dir)
        empty_profile_renderer = dbt.config.renderer.ProfileRenderer(generate_base_context({}))
        dbt_profile = dbt.config.Profile.from_raw_profiles(profiles_raw, profile_name, empty_profile_renderer)
        profile_data = extract_profile_data(profiles_raw, profile_name, dbt_profile.target_name)
        return dbt_profile.credentials, profile_data
    except DbtConfigError as exc:
        logger.debug(f"Failed parsing selected profile - {profiles_dir}, {profile_name}, {exc}")
        raise ConfigError(f"Failed parsing selected profile - {profiles_dir}, {profile_name}")


def get_bigquery_client(profile_credentials):
    if profile_credentials.impersonate_service_account:
        creds = \
            BigQueryConnectionManager.get_impersonated_bigquery_credentials(profile_credentials)
    else:
        creds = BigQueryConnectionManager.get_bigquery_credentials(profile_credentials)

    database = profile_credentials.database
    location = getattr(profile_credentials, 'location', None)

    info = client_info.ClientInfo(user_agent=f'elementary')
    return google.cloud.bigquery.Client(
        database,
        creds,
        location=location,
        client_info=info,
    )



from typing import Optional

from google.cloud import bigquery
from lineage.bigquery_query import BigQueryQuery
from lineage.query import Query
from lineage.query_context import QueryContext
from lineage.query_history import QueryHistory
from lineage.utils import get_logger
from datetime import datetime

logger = get_logger(__name__)


class BigQueryQueryHistory(QueryHistory):
    INFORMATION_SCHEMA_QUERY_HISTORY = """
    SELECT query, end_time, dml_statistics.inserted_row_count + dml_statistics.updated_row_count, statement_type, 
    user_email, destination_table, referenced_tables
           
    FROM region-{location}.INFORMATION_SCHEMA.JOBS_BY_PROJECT
    WHERE
         project_id = @project_id
         AND creation_time BETWEEN @start_time AND {creation_time_range_end_expr}
         AND end_time BETWEEN @start_time AND {end_time_range_end_expr}
         AND job_type = "QUERY"
         AND state = "DONE"
         AND error_result is NULL
         AND query NOT like '%JOBS_BY_PROJECT%'
    ORDER BY end_time
    """
    INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP = 'CURRENT_TIMESTAMP()'
    INFO_SCHEMA_END_TIME_UP_TO_PARAMETER = '@end_time'

    def __init__(self, con, should_export_query_history: bool = True, ignore_schema: bool = False,
                 dataset: str = None) -> None:
        self.dataset = dataset
        super().__init__(con, should_export_query_history, ignore_schema)

    @classmethod
    def _build_history_query(cls, start_date: datetime, end_date: datetime, database_name: str, location: str) -> \
            (str, []):
        query_parameters = [bigquery.ScalarQueryParameter("project_id", "STRING", database_name),
                            bigquery.ScalarQueryParameter("start_time", "TIMESTAMP", start_date)]

        end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP
        creation_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_CURRENT_TIMESTAMP
        if end_date is not None:
            query_parameters.append(bigquery.ScalarQueryParameter("end_time",
                                                                  "TIMESTAMP",
                                                                  cls._include_end_date(end_date)))
            end_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_PARAMETER
            creation_time_range_end_expr = cls.INFO_SCHEMA_END_TIME_UP_TO_PARAMETER

        query = cls.INFORMATION_SCHEMA_QUERY_HISTORY.format(location=location,
                                                            creation_time_range_end_expr=
                                                            creation_time_range_end_expr,
                                                            end_time_range_end_expr=
                                                            end_time_range_end_expr)
        return query, query_parameters

    def _query_history_table(self, start_date: datetime, end_date: datetime) -> [Query]:
        database_name = self.get_database_name()
        query, query_parameters = self._build_history_query(start_date, end_date, database_name, self._con.location)

        job_config = bigquery.QueryJobConfig(
            query_parameters=query_parameters
        )

        job = self._con.query(query, job_config=job_config)

        logger.debug("Finished executing bigquery jobs history query")

        queries = []
        rows = job.result()
        for row in rows:
            query_context = QueryContext(query_time=row[1],
                                         query_volume=row[2],
                                         query_type=row[3],
                                         user_name=row[4],
                                         destination_table=row[5],
                                         referenced_tables=row[6])

            query = BigQueryQuery(raw_query_text=row[0],
                                  query_context=query_context,
                                  profile_database_name=database_name,
                                  profile_schema_name=self.get_schema_name())

            queries.append(query)
        logger.debug("Finished fetching bigquery history job results")

        return queries

    def get_database_name(self) -> str:
        return self._con.project

    def get_schema_name(self) -> Optional[str]:
        return self.dataset if not self._ignore_schema else None



from typing import Optional
from lineage.query_context import QueryContext
from sqllineage.exceptions import SQLLineageException
from lineage.table_resolver import TableResolver
from lineage.query import Query
from lineage.utils import get_logger

logger = get_logger(__name__)


class BigQueryQuery(Query):

    EMPTY_QUERY_TYPE = ''
    DROP_PREFIX = 'DROP'
    ALTER_PREFIX = 'ALTER'
    PLATFORM_TYPE = 'BIGQUERY'

    @staticmethod
    def from_dict(query_dict: dict):
        query_context = QueryContext.from_dict(query_dict.pop('query_context'))
        if 'platform_type' in query_dict:
            query_dict.pop('platform_type')
        return BigQueryQuery(**query_dict, query_context=query_context)

    @staticmethod
    def _parse_table_json_column(table_resolver: TableResolver, table_json_column: dict) -> Optional[str]:
        if table_json_column is None:
            return None

        project = table_json_column.get('project_id')
        dataset = table_json_column.get('dataset_id')
        table = table_json_column.get('table_id')

        if project is None or dataset is None or table is None:
            return None

        if table.startswith('anon'):
            return None

        return table_resolver.name_qualification(f'{project}.{dataset}.{table}')

    def parse(self, full_table_names: bool = False) -> None:
        try:
            table_resolver = TableResolver(profile_database_name=self._profile_database_name,
                                           profile_schema_name=self._profile_schema_name,
                                           full_table_names=full_table_names)

            target_table = self._parse_table_json_column(table_resolver, self._query_context.destination_table)
            source_tables = set()
            for referenced_table in self._query_context.referenced_tables:
                source_table = self._parse_table_json_column(table_resolver, referenced_table)
                source_tables.add(source_table)

            query_type = self.EMPTY_QUERY_TYPE
            if self._query_context.query_type is not None:
                query_type = self._query_context.query_type

            if query_type.startswith(self.DROP_PREFIX):
                self.dropped_tables.add(target_table)
            elif query_type.startswith(self.ALTER_PREFIX):
                _, _, self.renamed_tables, _ = \
                    self._parse_query_text(table_resolver, self._raw_query_text)
            else:
                self.source_tables = source_tables
                self.target_tables.add(target_table)
        except SQLLineageException as exc:
            logger.debug(f'SQLLineageException was raised while parsing this query -\n{self._raw_query_text}\n'
                         f'Error was -\n{exc}.')




from setuptools import setup, find_packages
from io import open
from os import path
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    package_requirements = f.read().split('\n')

install_requirements = [x.strip() for x in package_requirements if ('git+' not in x) and (not x.startswith('#')) and
                        (not x.startswith('-'))]

dependency_links = [x.strip().replace('git+', '') for x in package_requirements if 'git+' not in x]

setup(
    name='elementary-lineage',
    description='Presenting data lineage based on your data warehouse query history',
    version='0.0.13',
    packages=find_packages(),
    python_requires='>=3.6.2',
    entry_points='''
        [console_scripts]
        edl=lineage.main:main
    ''',
    author="Elementary",
    package_data={"": ["header.html"]},
    keyword="data, lineage, data lineage, data warehouse, DWH",
    long_description=README,
    install_requires=install_requirements + [
        'jinja2<3.0.0',
        'werkzeug<2.0',
        'urllib3<1.26'
    ],
    long_description_content_type="text/markdown",
    license='',
    url='https://github.com/elementary-data/elementary-lineage',
    dependency_links=dependency_links,
    author_email='or@elementary-data.com',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)


from lineage.bigquery_query_history import BigQueryQueryHistory
from lineage.dbt_utils import extract_credentials_and_data_from_profiles, get_bigquery_client
from lineage.exceptions import ConfigError
from lineage.query_history import QueryHistory
from lineage.snowflake_query_history import SnowflakeQueryHistory
import snowflake.connector

snowflake.connector.paramstyle = 'numeric'


class QueryHistoryFactory(object):

    def __init__(self, profiles_dir: str, profile_name: str, export_query_history: bool,
                 ignore_schema: bool = False) -> None:
        self._profiles_dir = profiles_dir
        self._profile_name = profile_name
        self._export_query_history = export_query_history
        self._ignore_schema = ignore_schema

    def create_query_history(self) -> QueryHistory:
        credentials, profile_data = extract_credentials_and_data_from_profiles(self._profiles_dir, self._profile_name)
        credentials_type = credentials.type
        if credentials_type == 'snowflake':
            snowflake_con = snowflake.connector.connect(
                account=credentials.account,
                user=credentials.user,
                database=credentials.database,
                schema=credentials.schema,
                warehouse=credentials.warehouse,
                role=credentials.role,
                autocommit=True,
                client_session_keep_alive=credentials.client_session_keep_alive,
                application='elementary',
                **credentials.auth_args()
            )

            return SnowflakeQueryHistory(snowflake_con, self._export_query_history, self._ignore_schema,
                                         profile_data.get('query_history_source'))
        elif credentials_type == 'bigquery':
            bigquery_client = get_bigquery_client(credentials)
            return BigQueryQueryHistory(bigquery_client, self._export_query_history, self._ignore_schema,
                                        profile_data.get('dataset'))
        else:
            raise ConfigError("Unsupported profile type")
